@@ -4,9 +4,11 @@
 
 import os, re, sys, json, codecs, logging, requests, arrow, shutil, inspect, sqlalchemy, unicodedata, uuid, base64, datetime, time, pytz, xmltodict
 
+if os.path.dirname(sys.argv[0]) == '.':
+	sys.path.insert(0, '..')
+
 # https://github.com/serkanyersen/underscore.py/
 # pip3 install underscore.py
-
 from underscore import _
 
 from time import mktime
@@ -40,16 +42,15 @@ from sqlalchemy import not_
 
 from KnowledgeBase.model import *
 
-#=====================================================
 args = Argue()
 squirrel = Squirrel()
 logger = Logger()
 colours = Colours()
 treeify = Treeify(colour=True)
 
-#_____________________________________________________
 def quietly():
 	for logger in [
+		"sqlalchemy.engine.Engine",
 		"sqlalchemy.orm.relationships.RelationshipProperty",
 		"sqlalchemy.orm.strategies.LazyLoader", "sqlalchemy.orm.path_registry",
 		"sqlalchemy.orm.mapper.Mapper", "sqlalchemy.engine.base.Engine",
@@ -60,12 +61,10 @@ def quietly():
 		logging.getLogger(logger).setLevel(logging.WARNING)
 
 
-#_____________________________________________________
 def clean(text):
 	return unicodedata.normalize('NFKD', text).encode('ascii', 'ignore')
 
 
-#_____________________________________________________
 def public(node):
 	tipe = type(node)
 	if tipe is list:
@@ -78,7 +77,6 @@ def public(node):
 	return node
 	
 
-#_____________________________________________________
 def uncomment(value):
 	comment = re.compile('(\<\!\-\-|\-\-\>)')
 	value = value.replace('\n', '%%%%%')
@@ -94,13 +92,10 @@ def uncomment(value):
 	return output.getvalue().replace('%%%%%', '\n')
 
 
-#=====================================================
 class Empty(object):
 
 	base64 = 'Empty.base64'
 	
-	
-	#.................................................
 	def touch(self, file, adt=None):
 		stat = os.stat(file)
 		(st_mode, st_ino, st_dev, st_nlink, st_uid, st_gid, st_size, st_atime,
@@ -113,7 +108,6 @@ class Empty(object):
 		return adt
 		
 	
-	#.................................................
 	def toDB(self, b64, db):
 		t = self.touch(b64)
 		with open(b64, 'r') as fi:
@@ -126,7 +120,6 @@ class Empty(object):
 		return
 	
 	
-	#.................................................
 	def fromDB(self, db, b64):
 		t = self._touch(db)
 		with open(db, 'rb') as fi:
@@ -139,7 +132,6 @@ class Empty(object):
 		return
 	
 		
-#=====================================================
 @args.command(single=True)
 class Export(object):
 	
@@ -163,7 +155,6 @@ class Export(object):
 		return
 
 
-	#.................................................
 	def __init__(self, verbose=False):
 
 		if self.klobber:
@@ -200,12 +191,10 @@ class Export(object):
 		self.__Session.configure(bind=self.engine)
 
 
-	#.................................................
 	def __del__(self):
 		self.errorlog.close()
 
 
-	#.................................................
 	@property
 	def engine(self):
 		'''
@@ -216,7 +205,7 @@ class Export(object):
 		if not self.__engine:
 			self.__engine = sqlalchemy.create_engine(self.url)
 			
-		self.inspector = reflection.Inspector.from_engine(self.__engine)
+		self.inspector = sqlalchemy.inspect(self.__engine)
 		
 		self.tables = dict() # table: [ column ]
 		for table in self.inspector.get_table_names():
@@ -227,7 +216,6 @@ class Export(object):
 		return self.__engine
 
 
-	#.................................................
 	def Session(self):
 		'''
 		return a new session object
@@ -235,7 +223,6 @@ class Export(object):
 		return self.__Session()
 
 
-	#.................................................
 	def _sequences(self, session):
 		results = dict()
 		for sequence in session.query(t_sqlite_sequence):
@@ -243,7 +230,6 @@ class Export(object):
 		prettyPrintLn(results,align=True)	
 			
 
-	#.................................................
 	def _undone(self, session):
 		rows = list()
 		for undo in session.query(UndoLog):
@@ -251,16 +237,22 @@ class Export(object):
 		prettyPrintLn(rows)
 		
 	
-	#.................................................
 	def _process(self, outlines, session=None, parent=None, indent=''):
+		url_pattern=re.compile('.*(https*://\S+)\s*.*')
 		if type(outlines) is not list:
 			outlines = [outlines]
 		for outline in outlines:
 			item = Item()
 			session.add(item)
 			if '@text' in outline.keys():
-				item.Name = outline['@text']
-				print('%s%s'%(indent, outline['@text']))
+				text = outline['@text']
+				url_match = url_pattern.match(text)
+				if url_match:
+					url = url_match.group(1)
+					item.DirectLinkURL = url
+					text = text.replace(url,'')
+				item.Name = text
+				print('%s%s'%(indent, text))
 			if '@_note' in outline.keys():
 				item.Description = outline['@_note']
 				print('  %s"%s"'%(indent, outline['@_note']))
@@ -274,7 +266,6 @@ class Export(object):
 				self._process(outline['outline'], session=session, parent=item, indent='  %s'%indent)
 			
 					
-	#.................................................
 	@args.operation
 	@args.parameter(name='klobber', short='k', flag=True, help='create from scratch')
 	def load_opml(self, file, klobber=False):
@@ -286,8 +277,7 @@ class Export(object):
 		session.add(root)
 
 		root.Description = file
-		
-		opml = xmltodict.parse(open(file).read())
+		opml = xmltodict.parse(open(file,'rb').read().decode('UTF8'))
 		root.Name = opml['opml']['head']['title']
 
 		self._process(opml['opml']['body']['outline'], session=session, parent=root)
@@ -296,7 +286,6 @@ class Export(object):
 		session.close()
 
 		
-	#.................................................
 	@args.operation
 	def load_excel(self, file):
 		session = self.Session()
@@ -364,7 +353,6 @@ class Export(object):
 		session.close()			
 		
 		
-	#.................................................
 	@args.operation(name="import")
 	def importer(self):
 		session = self.Session()
@@ -393,7 +381,6 @@ class Export(object):
 		session.close()
 
 
-	#.................................................
 	@args.operation
 	@args.parameter(name='name', short='n', help='the name of the element')
 	@args.parameter(name='references', short='r', flag=True, help='list references')
@@ -402,6 +389,12 @@ class Export(object):
 	def inspect(self, name=None, references=False, categories=False, attachments=False):
 		session = self.Session()
 
+		if references:
+			query = session.query(Relation)
+			for reference in query.all():
+				print(f'{reference.inbound.Name} -> {reference.outbound.Name} : {reference.Name}')
+			return
+		
 		if categories:
 			query = session.query(ItemTemplate)
 			if name: query = query.filter_by(Name=name)
@@ -428,7 +421,6 @@ class Export(object):
 		session.close()
 
 
-	#.................................................
 	def _sort(self, session, item, depth=1, indent=''):
 		sys.stdout.write('%s\n'%(item.Name))
 		if depth > 0:
@@ -443,7 +435,6 @@ class Export(object):
 				self._sort(session, relation.outbound, depth-1, indent='%s\t'%indent)
 
 				
-	#.................................................
 	@args.operation
 	@args.parameter(name='name', help='name of node to sort references for')
 	@args.parameter(name='depth', short='d', type=int, help='recursion depth')
@@ -459,21 +450,23 @@ class Export(object):
 		session.close()
 
 		
-	#.................................................
 	@args.operation
 	@args.parameter(name='value', help='string to search in name for')
-	@args.parameter(name='description', short='d', flag=True, help='search in description as well')
-	def query(self, value, description=False):
+	@args.parameter(name='categories', short='c', nargs='+',  help='restrict to categories')
+	def query(self, value, description=False, categories=[]):
 		session = self.Session()
-		for item in session.query(Item):
-			if value in item.Name:
-				print(item.Name)
-			if description and value in item.Description:
-				print('\t%s'%item.Description)
+		
+		query = session.query(Item).filter(Item.Name.like(value)).join(Item.category)
+
+		for item in query.all():
+			if categories:
+				if item.category.TemplateName not in categories:
+					continue
+			print(item.Name)
+				
 		session.close()
 
 		
-	#.................................................
 	@args.operation
 	@args.parameter(name='old', help='string to search in name for')
 	@args.parameter(name='new', help='string to replace with')
@@ -498,7 +491,6 @@ class Export(object):
 		session.close()
 
 		
-	#.................................................
 	@args.operation
 	def clean(self):
 		session = self.Session()
@@ -516,18 +508,16 @@ class Export(object):
 		session.close()
 
 
-	#.................................................
 	@args.operation
 	def export(self):
 		shutil.copy2(self.database, '%s/.kdb'%self.exportdir)
-		self.inspector = reflection.Inspector.from_engine(self.engine)
+		self.inspector =  sqlalchemy.inspect(self.__engine)
 		for table in self.inspector.get_table_names():
 			if table.startswith('general_db'):
 				continue
 			self._table(table)
 
 
-	#.................................................
 	@args.operation
 	def deArrow(self):
 		session = self.Session()
@@ -543,7 +533,6 @@ class Export(object):
 		session.close()
 
 
-	#.................................................
 	def _column(self, column, value):
 		if not value: return ''
 		if not 'type' in column.keys(): return ''
@@ -574,7 +563,6 @@ class Export(object):
 		return '"%s"' % value
 
 
-	#.................................................
 	def _row(self, output, columns, row):
 		values = []
 		for c in range(len(row)):
@@ -586,7 +574,6 @@ class Export(object):
 		output.write('%s\n' % ('\t'.join(values)))
 
 
-	#.................................................
 	def _header(self, table):
 		columns = dict(
 			map(lambda x: (x['name'], x), self.inspector.get_columns(table)))
@@ -600,7 +587,6 @@ class Export(object):
 		return header
 
 
-	#.................................................
 	def _table(self, table):
 		print(table)
 		file = '%s/%s.csv' % (self.exportdir, table)
@@ -623,11 +609,16 @@ class Export(object):
 				print(input.read())
 
 
-#=====================================================
 def main():
 	quietly()
+	if False:
+		args.parse([
+			'-D',
+			'libraries.kdb',
+			'load_opml',
+			'libraries.opml',
+		])
 	args.execute()
 
 		
-#_____________________________________________________
 if __name__ == '__main__': main()
